import networkx as nx
import numpy as np

from ...utils import soft_eccentricity
from ..source_results import SingleSourceResult


def jordan_centrality(I, G):
    """Computes the infection eccentricity of each node in the infection subgraph. To
    produce a score with highest value corresponding to the Jordan center, we return
    the inverse of the infection eccentricity.

    Parameters
    ----------
    I : NetworkX Graph
        The infection subgraph observed at a particular time step
    G : NetworkX Graph
        The original graph the infection process was run on.
        I is a subgraph of G induced by infected vertices at observation time.

    Notes
    -----
    The Jordan infection center is the vertex with minimum infection eccentricity.
    This is described in [1]_ and [2]_.

    Nodes outside the infection subgraph receive a score of negative infinity.


    Examples
    --------
    >>> result = cosasi.single_source.jordan_centrality(I, G)

    References
    ----------
    .. [1] L. Ying and K. Zhu,
        "On the Universality of Jordan Centers for Estimating Infection Sources in Tree Networks"
        IEEE Transactions of Information Theory, 2017
    .. [2] L. Ying and K. Zhu,
        "Diffusion Source Localization in Large Networks"
        Synthesis Lectures on Communication Networks, 2018
    """
    scores = {
        v: 1 / soft_eccentricity(I, v=v) if v in I.nodes else -np.inf for v in G.nodes
    }
    result = SingleSourceResult(
        source_type="single-source",
        inference_method="jordan centrality",
        scores=scores,
        G=G,
        reverse=True,
    )
    return result



import math
import random
import warnings

import scipy
import numpy as np
import networkx as nx
from sklearn.cluster import SpectralClustering

from .helpers import attack_degree, attack_degree_partition
from ..source_inference.multiple_source import netsleuth


def source_subgraphs(I, number_sources=2):
    """Subdivides the provided graph into specified number of subgraphs
    via spectral clustering.

    Parameters
    ----------
    I : NetworkX Graph
        The infection subgraph observed at a particular time step
    number_sources : int
        The hypothesized number of infection sources
    """
    warnings.filterwarnings("ignore", category=UserWarning)
    warnings.filterwarnings("ignore", category=FutureWarning)

    I_node_list = list(I.nodes)
    A = nx.adjacency_matrix(I)
    subgraphs = []
    sc = SpectralClustering(number_sources, affinity="precomputed", n_init=100)
    sc.fit(scipy.sparse.csr_matrix(A))
    subgraph_labels = sc.labels_
    unique_subgraph_labels = set(subgraph_labels)
    for i in unique_subgraph_labels:
        subgraph_nodes = [I_node_list[j] for j in np.where(subgraph_labels == i)[0]]
        subgraphs.append(I.subgraph(subgraph_nodes))
    return subgraphs


def number_sources(
    I,
    number_sources=None,
    return_source_subgraphs=True,
    number_sources_method="eigengap",
    G=None,
):
    """Manages source subgraph estimation, mostly via spectral analysis and clustering.

    Parameters
    ----------
    I : NetworkX Graph
        The infection subgraph observed at a particular time step
    number_sources : int or None (optional)
        if int, this is the hypothesized number of infection sources
        if None, estimates the number of sources via Eigengap heuristic
    return_source_subgraphs : bool
        if True, returns subgraphs of I corresponding to each hypothesized infection source
        if False, does not return subgraphs of I corresponding to each hypothesized infection source
    number_sources_method : str
        method for estimating the number of sources. one of the following options:
        - "eigengap" : uses the Eigengap of the normalized graph Laplacian to estimate the number of clusters
        - "netsleuth" : runs the multi-source NETSLEUTH algorithm and reports the number of seeds
        - "chatter" : invokes a spectral method based on the Chatter algorithm
        if number_sources != None, this doesn't do anything
    G : NetworkX Graph (optional)
        the original network the contagion process was run on
        generally optional (e.g. not needed for eigengap), occassionally required (e.g. needed for netsleuth)

    Notes
    -----
    If the diffusion process is brief or observation is early, and infection sources
    are sufficiently sparse, then the infected subgraphs corresponding to each infection
    source may be the connected components of the input graph. This is described in
    Section 2.6 of [1]_.

    We estimate the number of infection sources by the minimum of the number of connected
    components and the Eigengap heuristic of the provided graph. The Eigengap heuristic
    is described in [2]_.

    With a hypothesized number of infection sources in hand, we partition the graph via
    spectral clustering to provide a list of subgraphs corresponding to each infection
    source [3]_.

    References
    ----------
    .. [1] L. Ying and K. Zhu,
        "Diffusion Source Localization in Large Networks"
        Synthesis Lectures on Communication Networks, 2018
    .. [2] U. von Luxburg,
        "A Tutorial on Spectral Clustering"
        Statistics and Computing, 2007
        https://link.springer.com/article/10.1007/s11222-007-9033-z
    .. [3] A. Damle and V. Minden and L. Ying
        "Simple, direct and efficient multi-way spectral clustering"
        Information and Inference: A Journal of the IMA, 2019
        https://academic.oup.com/imaiai/article/8/1/181/5045955
    """
    if isinstance(number_sources, int):
        if return_source_subgraphs:
            return number_sources, source_subgraphs(I, number_sources=number_sources)
        else:
            return number_sources
    elif isinstance(number_sources, type(None)):
        if number_sources_method.lower() == "eigengap":
            m = eigengap(I)
        elif number_sources_method.lower() == "netsleuth":
            if isinstance(G, type(None)):
                raise ValueError("Need `G` for NETSLEUTH method.")
            netsleuth_result = netsleuth(I=I, G=G, hypotheses_per_step=1)
            m = len(netsleuth_result.topn(1)[0])
        elif number_sources_method.lower() == "chatter":
            if isinstance(G, type(None)):
                raise ValueError("Need `G` for chatter method.")
            m = chatter(I, G)
        else:
            raise NotImplementedError

        if m <= nx.number_connected_components(I):
            subgraphs = [I.subgraph(c) for c in nx.connected_components(I)]
            m = len(subgraphs)
        else:
            subgraphs = source_subgraphs(I, number_sources=m)

        if return_source_subgraphs:
            return m, subgraphs
        else:
            return m
    else:
        raise ValueError("number_sources not recognized: must be an integer or None.")


def chatter(I, G):
    """Estimates the number of sources of a graph diffusion process via the Chatter algorithm.

    Parameters
    ----------
    I : NetworkX Graph
        The infection subgraph observed at a particular time step
    G : NetworkX Graph
        The graph the diffusion process was originally run on
    """
    # T = list(I.nodes)
    # S = [v for v in G if v not in T]
    # frontier = nx.node_boundary(G=G, nbunch1=S, nbunch2=T)
    # frontier_idx = [T.index(v) for v in frontier]
    freq = chatter_frequency(I)
    np.fill_diagonal(freq, 0)
    w, v = np.linalg.eig(freq)
    return int(np.argmax((1 / (w + 1))[1 : len(I)]) + 1)


def eigengap(G):
    """Returns the estimated number of clusters of G, based on the Eigengap
    of the normalized graph Laplacian.

    Parameters
    ----------
    G : NetworkX Graph
        The graph to analyze

    Notes
    -----
    The Eigengap heuristic is described in [1]_.

    References
    ----------
    .. [1] U. von Luxburg,
        "A Tutorial on Spectral Clustering"
        Statistics and Computing, 2007
        https://link.springer.com/article/10.1007/s11222-007-9033-z
    """
    warnings.filterwarnings("ignore", category=FutureWarning)

    L = nx.normalized_laplacian_matrix(G).toarray()
    eigenvalues, eigenvectors = np.linalg.eig(L)
    eigenvalues.sort()
    k = np.argmax(np.diff(eigenvalues)) + 1
    return k


def bits_encode_integer(n):
    """Estimates the number of bits required to encode an integer n>=1.

    Parameters
    ----------
    n : int
        an integer at least 1

    Notes
    -----
    Calculation is from Section 4.1 of [1]_.

    References
    ----------
    .. [1] B. A. Prakash, J. Vreeken, C. Faloutsos,
        "Efficiently spotting the starting points of an epidemic in a large graph"
        Knowledge and Information Systems, 2013
        https://link.springer.com/article/10.1007/s10115-013-0671-5
    """
    if n < 1:
        raise ValueError("n must be at least 1")
    l = math.log(2.865064)
    c = math.log(n)
    while c > 0:
        l += c
        c = math.log(c)
    return l


def bits_encode_seed(s, G):
    """Number of bits required to identify a seed set (hypothesized
    infection source set).

    Parameters
    ----------
    s : array-like
        seed set
    G : NetworkX Graph
        The original graph the infection process was run on.

    Notes
    -----
    Calculation is from Section 4.1 of [1]_.

    References
    ----------
    .. [1] B. A. Prakash, J. Vreeken, C. Faloutsos,
        "Efficiently spotting the starting points of an epidemic in a large graph"
        Knowledge and Information Systems, 2013
        https://link.springer.com/article/10.1007/s10115-013-0671-5
    """
    n = len(s)
    return bits_encode_integer(n) + math.log(math.comb(len(G), n))


def bits_encode_ripple(s, G, beta=0.01):
    """Total description length of a seed set and its corresponding maximum likelihood
    propagation ripple.

    Parameters
    ----------
    s : array-like
        seed set
    G : NetworkX Graph
        The original graph the infection process was run on.
    beta : float
        infection probability

    Notes
    -----
    Calculation is from Section 4.3 of [1]_.

    References
    ----------
    .. [1] B. A. Prakash, J. Vreeken, C. Faloutsos,
        "Efficiently spotting the starting points of an epidemic in a large graph"
        Knowledge and Information Systems, 2013
        https://link.springer.com/article/10.1007/s10115-013-0671-5
    """

    def probability_infection(m_d, f_d, d):
        """Probability of m_d nodes being infected in a subset of the frontier.

        Parameters
        ----------
        m_d : int
            number of nodes infected
        f_d :
            number of nodes in a frontier subset F_d
        d : int
            degree

        Notes
        -----
        Calculation is from Section 4.2.3 of [1]_.

        References
        ----------
        .. [1] B. A. Prakash, J. Vreeken, C. Faloutsos,
            "Efficiently spotting the starting points of an epidemic in a large graph"
            Knowledge and Information Systems, 2013
            https://link.springer.com/article/10.1007/s10115-013-0671-5
        """
        p_d = 1 - (1 - beta) ** d  # attack probability in the set
        return math.comb(f_d, m_d) * (p_d**m_d) * (1 - p_d) ** (f_d - m_d)

    def l_frontier(f, infected=s):
        """Calculates the code length for encoding the infectious in the frontier
        set at a snapshot of time.

        Parameters
        ----------
        f : array-like
            frontier set
        infected : array-like
            infected nodes

        Notes
        -----
        Calculation is Equation 3 from Section 4.2.4 of [1]_.

        References
        ----------
        .. [1] B. A. Prakash, J. Vreeken, C. Faloutsos,
            "Efficiently spotting the starting points of an epidemic in a large graph"
            Knowledge and Information Systems, 2013
            https://link.springer.com/article/10.1007/s10115-013-0671-5
        """
        l = 0
        partition = attack_degree_partition(f, infected, G)
        for d in partition:
            f_d = len(partition[d])
            m_d = int(min(math.floor(p_d * (f_d + 1)), f_d))
            if m_d == 0 or f_d == 0:
                continue
            l -= (
                math.log(probability_infection(m_d, f_d, d))
                + m_d * math.log(m_d / f_d)
                + (f_d - m_d) * math.log(1 - m_d / f_d)
            )
        return l

    infected = s
    frontier = set([j for i in infected for j in G.neighbors(i) if j not in infected])
    bits_ripple = 0
    t = 0  # index starts at 0 per p. 42 / Section 4.2.2
    while len(frontier) > 0 and len(infected) < len(G):
        # ripple step, get new frontier
        partition = attack_degree_partition(frontier, infected, G)
        for d in partition:
            f_d = len(partition[d])
            p_d = 1 - (1 - beta) ** d  # attack probability in the set
            n_d = math.floor((f_d / beta + 1) * p_d)

            infected += random.sample(partition[d], min(n_d, f_d))
        frontier = set(
            [j for i in infected for j in G.neighbors(i) if j not in infected]
        )
        infected = list(set(infected))
        bits_ripple += l_frontier(frontier, infected)
        t += 1
    return bits_encode_integer(t) + bits_ripple


def description_length(s, G, beta=0.01):
    """Implements a greedy heuristic to estimate the two-part minimal infection
    description length of a proposed set of infection sources.

    Parameters
    ----------
    s : array-like
        seed set
    G : NetworkX Graph
        The original graph the infection process was run on.
    beta : float
        infection probability

    Notes
    -----
    The minimal description length, as applied to source localization, is introduced
    in [1]_.

    References
    ----------
    .. [1] B. A. Prakash, J. Vreeken, C. Faloutsos,
        "Efficiently spotting the starting points of an epidemic in a large graph"
        Knowledge and Information Systems, 2013
        https://link.springer.com/article/10.1007/s10115-013-0671-5
    """
    return bits_encode_seed(s, G) + bits_encode_ripple(s=s, G=G, beta=0.01)


def chatter_frequency(G, t=None):
    """Implements the Chatter Algorithm described in Notes.

    Parameters
    ----------
    G : NetworkX Graph
        The graph to analyze
    t : int or None (optional)
        number of rounds to complete
        if None, the algorithm runs until every node's message is received by
        every other node at least 5 times.

    Notes
    -----
    Each node starts with a message bank consisting of its own ID.
    For `t` many rounds, each node broadcasts its message bank to its neighbors,
    and all nodes receiving messages append them to their own message bank.
    message_frequency[i][j] is the number of times i received j's message.

    A "naive"/pure message-passing formulation of this would be along the lines of:

    .. code-block:: python

        def chatter_distance_slow(G, t):
            messages = {i:[i] for i in G}
            for _ in range(t):
                new_messages = copy.deepcopy(messages)
                for i in range(len(G)):
                    for j in G.neighbors(i):
                        new_messages[j] += messages[i]
                messages = new_messages
            return messages

    where messages[i].count(j) is the number of times i received j's message. But
    this is very slow and easily re-written as matrix multiplication, as is done
    here.
    """
    warnings.filterwarnings("ignore", category=FutureWarning)

    A = nx.adjacency_matrix(G).toarray()
    message_frequency = scipy.sparse.identity(len(G)).toarray()
    if isinstance(t, type(None)):
        if not nx.is_connected(G):
            return chatter_frequency(G, t=len(G))

        while np.min(message_frequency) < 5:
            for i in range(len(G)):
                message_frequency[i] += A.dot(message_frequency[i])
    else:
        for _ in range(t):
            for i in range(len(G)):
                message_frequency[i] += A.dot(message_frequency[i])
    return message_frequency


def chatter_distance(G, t, u=None, v=None, normalized=True):
    """Invokes the Chatter Algorithm/chatter frequency to obtain chatter distance,
    a graph topology metric.

    Parameters
    ----------
    G :NetworkX Graph
        The graph to analyze
    t : int
        number of rounds to complete
    u : node (optional)
        starting node. if not provided, we return an array of distances
    v : node (optional)
        end node. if not provided, we return an array of distances
    normalized : bool
        if True, all distances are scaled to have a max value of 1

    Notes
    -----
    The chatter distance between nodes `u` and `v` reflects the difficulty node `u`
    is expected to have in transmitting a message to node `v`.
    """
    message_frequency = chatter_frequency(G, t)
    distance = 1 / message_frequency
    if normalized:
        distance /= np.max(distance)
    if isinstance(u, type(None)) and isinstance(v, type(None)):
        return distance
    if isinstance(v, type(None)):
        return distance[u]
    if isinstance(u, type(None)):
        return distance[:][v]
    return distance[u][v]



import networkx as nx
import numpy as np
import warnings

from ..source_results import SingleSourceResult


def netsleuth(I, G):
    """Implements the single-source NETSLEUTH algorithm to score all nodes in G.

    Parameters
    ----------
    I : NetworkX Graph
        The infection subgraph observed at a particular time step
    G : NetworkX Graph
        The original graph the infection process was run on.
        I is a subgraph of G induced by infected vertices at observation time.

    Notes
    -----
    NETSLEUTH is described in [1]_. General idea is that, under mean field
    approximation, the probability of observing an infection subgraph given a
    particular source s is proportional to the sth entry of the largest eigenvector
    of the infection subgraph Laplacian. The implementation below is described in
    [2]_.

    Nodes outside the infection subgraph (i.e. the frontier set) receive a score of
    negative infinity.

    NETSLEUTH has linear complexity with the number of edges of the infected subgraph,
    edges of the frontier set, and vertices of the infected subgraph.


    Examples
    --------
    >>> result = cosasi.single_source.netsleuth(I, G)

    References
    ----------
    .. [1] B. A. Prakash, J. Vreeken, C. Faloutsos,
        "Efficiently spotting the starting points of an epidemic in a large graph"
        Knowledge and Information Systems, 2013
        https://link.springer.com/article/10.1007/s10115-013-0671-5
    .. [2] L. Ying and K. Zhu,
        "Diffusion Source Localization in Large Networks"
        Synthesis Lectures on Communication Networks, 2018
    """
    warnings.filterwarnings("ignore", module="networkx\..*")

    L = nx.laplacian_matrix(G).toarray()
    infection_indices = [i for i in I.nodes]
    L_I = L[np.ix_(infection_indices, infection_indices)]
    eigenvalues, eigenvectors = np.linalg.eig(L_I)
    largest_eigenvalue = max(eigenvalues)
    largest_eigenvector = eigenvectors[:, list(eigenvalues).index(largest_eigenvalue)]
    scores = {
        v: largest_eigenvector[infection_indices.index(v)]
        if v in infection_indices
        else -np.inf
        for v in G.nodes
    }
    result = SingleSourceResult(
        source_type="single-source", inference_method="netsleuth", scores=scores, G=G
    )
    return result



import torch_geometric.datasets as ds
import random
import ndlib.models.epidemics as ep
import ndlib.models.ModelConfig as mc


def SI(g, config, seed, rounds=100, beta=0.1):

    result = []

    for iter in range(rounds):

        model_temp = ep.SIModel(g) # _temp
        config_temp = mc.Configuration()
        config_temp.add_model_initial_configuration('Infected', seed)
        config_temp.add_model_parameter('beta', beta)

        for a, b in g.edges(): # _temp
            weight = config.config["edges"]['threshold'][(a, b)]
            config_temp.add_edge_configuration('threshold', (a, b), weight)

        model_temp.set_initial_status(config_temp)

        iterations = model_temp.iteration_bunch(5)

        result.append(iterations[4]['node_count'][1])

    return result



import networkx as nx
import numpy as np
import torch_geometric.datasets as ds
import random
import ndlib
import ndlib.models.epidemics as ep
import ndlib.models.ModelConfig as mc

from torch_geometric.datasets import Planetoid

def CiteSeer():
    dataset = Planetoid(root='./Planetoid', name='CiteSeer')  # Cora, CiteSeer, PubMed
    data = dataset[0]
    edges = (data.edge_index.numpy()).T.tolist()
    G = nx.from_edgelist(edges)

    c = max(nx.connected_components(G), key=len)
    g = G.subgraph(c).copy()
    config = mc.Configuration()

    for a, b in g.edges():
        weight = random.randrange(40,80)
        weight = round(weight / 100, 2)
        config.add_edge_configuration("threshold", (a, b), weight)
        g[a][b]['weight'] = weight

    return g, config

def PubMed():
    dataset = Planetoid(root='./Planetoid', name='PubMed')  # Cora, CiteSeer, PubMed
    data = dataset[0]
    edges = (data.edge_index.numpy()).T.tolist()
    G = nx.from_edgelist(edges)

    c = max(nx.connected_components(G), key=len)
    g = G.subgraph(c).copy()
    config = mc.Configuration()

    for a, b in g.edges():
        weight = random.randrange(40,80)
        weight = round(weight / 100, 2)
        config.add_edge_configuration("threshold", (a, b), weight)
        g[a][b]['weight'] = weight

    return g, config

def Cora():
    dataset = Planetoid(root='./Planetoid', name='Cora')  # Cora, CiteSeer, PubMed
    data = dataset[0]
    edges = (data.edge_index.numpy()).T.tolist()
    G = nx.from_edgelist(edges)

    c = max(nx.connected_components(G), key=len)
    g = G.subgraph(c).copy()
    config = mc.Configuration()

    for a, b in g.edges():
        weight = random.randrange(40,80)
        weight = round(weight / 100, 2)
        config.add_edge_configuration("threshold", (a, b), weight)
        g[a][b]['weight'] = weight

    return g, config

def photo():
    dataset = ds.Amazon(root='./geo', name = 'Photo')
    data = dataset[0]
    edges = (data.edge_index.numpy()).T.tolist()
    G = nx.from_edgelist(edges)
    g = nx.convert_node_labels_to_integers(G, first_label=0, ordering='default', label_attribute=None)

    config = mc.Configuration()

    for a, b in g.edges():
        weight = random.randrange(5,20)
        weight = round(weight / 100, 2)
        config.add_edge_configuration("threshold", (a, b), weight)
        g[a][b]['weight'] = weight

    return g, config

def coms():
    dataset = ds.Amazon(root='./geo', name = 'Computers')
    data = dataset[0]
    edges = (data.edge_index.numpy()).T.tolist()
    G = nx.from_edgelist(edges)
    g = nx.convert_node_labels_to_integers(G, first_label=0, ordering='default', label_attribute=None)

    config = mc.Configuration()

    for a, b in g.edges():
        weight = random.randrange(5,20)
        weight = round(weight / 100, 2)
        config.add_edge_configuration("threshold", (a, b), weight)
        g[a][b]['weight'] = weight

    return g, config



import random
import numpy as np
import operator

import networkx as nx
import ndlib.models.epidemics as ep
import ndlib.models.ModelConfig as mc


class StaticNetworkContagion:
    """A stochastic epidemic process defined on a static network.

    Parameters
    ----------
    G : NetworkX Graph
        The network for the diffusion process to run on
    model : str
        Specifies the epidemic model. Currently handles the following diffusion models:
            - SI (susceptible-infected)
            - SIS (susceptible-infected-susceptible)
            - SIR (susceptible-infected-recovered)
    infection_rate : float
        Inter-node infection efficiency
        must be in [0, 1]
    recovery_rate : float or None
        The recovery rate
        must be in [0, 1] (or None if diffusion model is SI)
    fraction_infected : float or None
        fraction of nodes to initialize as infected (selected uniformly at random)
        if both fraction_infected and number_infected are None, initializes with 1 infected node
    number_infected : float or None
        number of nodes to initialize as infected (selected uniformly at random)
        if both fraction_infected and number_infected are None, initializes with 1 infected node
    seed : integer, random_state, or None (default)
        random number generation state.

    Notes
    -----
    A wrapper for `ndlib` with convenience utilities added.
    """

    def __init__(
        self,
        G,
        model="si",
        infection_rate=0.01,
        recovery_rate=None,
        fraction_infected=None,
        number_infected=None,
        seed=None,
        model_config=None # added by Zhiqian
    ):
        """A stochastic epidemic process defined on a static network.

        Parameters
        ----------
        G : NetworkX Graph
            The network for the diffusion process to run on
        model : str
            Specifies the epidemic model. Currently handles the following diffusion models:
                SI
                SIS
                SIR
        infection_rate : float
            Inter-node infection efficiency
            must be in [0, 1]
        recovery_rate : float or None
            The recovery rate
            must be in [0, 1] (or None if diffusion model is SI)
        fraction_infected : float or None
            fraction of nodes to initialize as infected (selected uniformly at random)
            if both fraction_infected and number_infected are None, initializes with 1 infected node
        number_infected : float or None
            number of nodes to initialize as infected (selected uniformly at random)
            if both fraction_infected and number_infected are None, initializes with 1 infected node
        seed : integer, random_state, or None (default)
            random number generation state.

        Notes
        -----
        A wrapper for `ndlib` with convenience utilities added.
        """
        self.model = model.lower()
        self.seed = seed
        self.model_config = model_config if model_config else mc.Configuration()  # added by Zhiqian


        if not isinstance(self.seed, type(None)):
            random.seed(self.seed)
            np.random.seed(self.seed)

        if isinstance(G, nx.classes.graph.Graph):
            self.G = G
        else:
            raise ValueError("G must be a NetworkX instance.")

        if isinstance(infection_rate, float) and 0.0 <= infection_rate <= 1.0:
            self.infection_rate = infection_rate
        else:
            raise ValueError("Infection rate must be a float between 0 and 1.")

        if not recovery_rate or (
            isinstance(recovery_rate, float) and 0.0 <= recovery_rate <= 1.0
        ):
            self.recovery_rate = recovery_rate
        else:
            raise ValueError("Recovery rate must be a float between 0 and 1.")

        if fraction_infected and number_infected:
            raise ValueError(
                "User can only provide one of fraction_infected, number_infected."
            )
        elif not fraction_infected and not number_infected:
            self.fraction_infected = fraction_infected
            self.number_infected = 1
        else:
            self.fraction_infected = fraction_infected
            self.number_infected = number_infected

        self._init_sim()
        self.history = []
        return None

    def _init_sim(self):
        """Initializes the diffusion process properties and initial infectivity."""        
        config = self.model_config # added by Zhiqian
        config.add_model_parameter("beta", self.infection_rate)

        if self.model == "sir":
            self.sim = ep.SIRModel(graph=self.G, seed=self.seed)
            if not self.recovery_rate:
                raise ValueError("Recovery rate must be defined for SIR model.")
            config.add_model_parameter("gamma", self.recovery_rate)
        elif self.model == "si":
            self.sim = ep.SIModel(graph=self.G, seed=self.seed)
        elif self.model == "sis":
            self.sim = ep.SISModel(graph=self.G, seed=self.seed)
            if not self.recovery_rate:
                raise ValueError("Recovery rate must be defined for SIS model.")
            config.add_model_parameter("lambda", self.recovery_rate)
        else:
            raise NotImplementedError("Diffusion model not recognized.")

        if self.number_infected:
            if not isinstance(self.seed, type(None)):
                random.seed(self.seed)
            infected = random.sample(range(len(self.G)), self.number_infected)
            config.add_model_initial_configuration("Infected", infected)
        elif self.fraction_infected:
            config.add_model_parameter("fraction_infected", self.fraction_infected)
        elif self.mc: # added by Zhiqian
            config = self.model_config
        else:
            raise NotImplementedError

        self.sim.set_initial_status(config)
        return None

    def forward(self, steps=100, verbose=False):
        """Executes specified number of diffusion process steps. Records simulation history.

        Parameters
        ----------
        steps : int
            Number of simulation steps.
        verbose : bool (default False)
            Specifies whether to return the simulation history.

        Notes
        -----
        Can be run more than once; this just adds steps to the simulation history.
        """
        self.history += self.sim.iteration_bunch(steps)
        if verbose:
            return self.history
        return None

    def reset_sim(self):
        """Resets the simulation to its initialized states. Does not preserve compartmental histories."""
        self.history = []
        self.sim.reset()
        return None

    def get_infected_indices(self, step=0):
        """Retrieves the indices of all vertices in the infected compartment at the provided step.

        Parameters
        ----------
        step : int
            Iteration step

        Returns
        -------
        list
        """
        nodes = list(self.G)

        def status_to_delta(status):
            """Converts the history's status to a vector representing movement in
            (+1) and out (-1) of the infected compartment

            Parameters
            ----------
            status : dict
                status dictionary from history, e.g. self.history[step]["status"]
            """
            delta = np.zeros(len(self.G))
            for idx in status:
                s = status[idx]
                if s == 1:
                    # node became infected this step
                    delta[idx] = 1
                if s == 2:
                    # node became removed this step
                    delta[idx] = -1
            return delta

        if step >= len(self.history):
            raise ValueError(
                "Invalid step. Continue the simulation to reach this step."
            )
        infected = np.zeros(len(self.G))
        for s in range(step + 1):
            infected += status_to_delta(self.history[s]["status"])
        return [nodes[i] for i in np.where(infected == 1)[0]]

    def get_infected_subgraph(self, step=0):
        """Returns the subgraph of the contact network whose vertices are marked infected.

        Parameters
        ----------
        step : int
            Iteration step

        Returns
        -------
        NetworkX Graph

        Notes
        -----
        This is only guaranteed to be connected in the SI model.
        """
        infected_indices = self.get_infected_indices(step=step)
        not_infected_indices = set(self.G.nodes) - set(infected_indices)
        H = self.G.copy()
        H.remove_nodes_from(not_infected_indices)
        return H

    def get_observers(self, observers=1):
        """Observers record the step number when they become infected. For a specified number
        or list of observers, returns a dict of observers and the timestamps at which they
        become infected.

        Parameters
        ----------
        observers : int or list
            If int, observers specifies the number of observation nodes
            If list, observers specifies the observation nodes directly

        Notes
        -----
        If self.model == "sis", nodes may be reinfected, so observers record a list of the timestamps
        at which they are infected. Otherwise, observers record one timestamp (step number) only.

        If an observer is not infected during the simulation history, its corresponding infection
        timestamp is recorded as infinity.
        """
        if not self.history:
            raise ValueError(
                "Simulation must be run before retrieving observer information."
            )
        timestamp_placeholder = np.inf if self.model == "si" else list()
        if isinstance(observers, int):
            if not isinstance(self.seed, type(None)):
                random.seed(self.seed)
                np.random.seed(self.seed)

            observer_dict = {
                i: timestamp_placeholder for i in random.sample(self.G.nodes, observers)
            }
        elif isinstance(observers, list):
            observer_dict = {i: timestamp_placeholder for i in observers}
        else:
            raise NotImplementedError

        for i in range(len(self.history)):
            status = self.history[i]["status"]
            if self.model == "sis":
                for j in observer_dict:
                    if j in status and status[j] == 1:
                        observer_dict[j].append(i)
            else:
                for j in observer_dict:
                    if j in status and status[j] == 1:
                        observer_dict[j] = i
        return observer_dict

    def get_source(self, return_subgraph=False):
        """Returns the vertices marked infected at initialization.

        Parameters
        ----------
        return_subgraph : bool
            If True, returns a subgraph of infected vertices.
            If False, returns a list of indices.

        Returns
        -------
        list or NetworkX Graph
        """
        if not isinstance(return_subgraph, bool):
            raise ValueError("return_subgraph param must be a bool")
        if return_subgraph:
            return self.get_infected_subgraph(step=0)
        return self.get_infected_indices(step=0)

    def get_frontier(self, step=0):
        """Retrieves the frontier set of a given step. This is the set of infected nodes
        with an uninfected neighbor.

        Parameters
        ----------
        step : int
            Iteration step

        Returns
        -------
        NetworkX Graph

        Notes
        -----
        In the SI model, the frontier set consists of nodes likely to have been
        infected last, by the given timestep.
        """
        T = self.get_infected_indices(step=step)
        S = [v for v in self.G if v not in T]
        frontier = nx.node_boundary(G=self.G, nbunch1=S, nbunch2=T)
        return frontier
    


from FlowTasks import forward, backward, graph_eval



# ### Testing / Examples

# a FW1 dataset, with the observations stored as attributes to a networkx graph

# In[23]:


output = forward(1, obs_type = 'networkx', num_results=5)

#print observation type
print('Observations are of type:', type(output[0]['observations'][0]['observation']), end='\n\n')

for result in output:
    observations = result['observations']
    graph = result['base_graph']
    sir = result['SIR_model']
    
    #print sir values for this result
    print('SIR model has values: ', sep='', end='')
    print(f'beta = {round(sir["beta"],3)}, gamma = {round(sir["gamma"],3)}')
    
    #print observation time intervals for this result
    print('Observations at time intervals: ', sep='', end='')
    for ss in observations:
        print(f'{ss["time"]}, ', end='')
    print()
    
    start = observations[0]
    for i in range(1, len(observations)):
        print(f'Predicting {observations[i]["time"]} is the same as {start["time"]} - ', end='')
        eval_dict = graph_eval(observations[i]["observation"], start["observation"])
        for key, value in eval_dict.items():
            print(f'{key}: {round(value,3)}, ', end='')
        print()
    print()


# a FW1_2 dataset, with the observations stored in numpy arrays

# In[24]:


output = forward([1,2], obs_type = 'numpy', num_results=5)

#print observation type
print('Observations are of type:', type(output[0]['observations'][0]['observation']), end='\n\n')

for result in output:
    observations = result['observations']
    graph = result['base_graph']
    sir = result['SIR_model']
    
    #print sir values for this result
    print('SIR model has values: ', sep='', end='')
    print(f'beta = {round(sir["beta"],3)}, gamma = {round(sir["gamma"],3)}')
    
    #print observation time intervals for this result
    print('Observations at time intervals: ', sep='', end='')
    for ss in observations:
        print(f'{ss["time"]}, ', end='')
    print()
    
    start = observations[0]
    for i in range(1, len(observations)):
        print(f'Predicting {observations[i]["time"]} is the same as {start["time"]} - ', end='')
        eval_dict = graph_eval(observations[i]["observation"], start["observation"])
        for key, value in eval_dict.items():
            print(f'{key}: {round(value,3)}, ', end='')
        print()
    print()


# a BW1_4 dataset, with the observations stored as pytorch geometric data objects

# In[25]:


output = backward([1,4], obs_type = 'torch', num_results=5)

#print observation type
print('Observations are of type:', type(output[0]['observations'][0]['observation']), end='\n\n')

for result in output:
    observations = result['observations']
    graph = result['base_graph']
    sir = result['SIR_model']
    
    #print sir values for this result
    print('SIR model has values: ', sep='', end='')
    print(f'beta = {round(sir["beta"],3)}, gamma = {round(sir["gamma"],3)}')
    
    #print observation time intervals for this result
    print('Observations at time intervals: ', sep='', end='')
    for ss in observations:
        print(f'{ss["time"]}, ', end='')
    print()
    
    start = observations[0]
    for i in range(1, len(observations)):
        print(f'Predicting {observations[i]["time"]} is the same as {start["time"]} - ', end='')
        eval_dict = graph_eval(observations[i]["observation"], start["observation"])
        for key, value in eval_dict.items():
            print(f'{key}: {round(value,3)}, ', end='')
        print()
    print()



"""Generic objects for the result of single-source and multi-source localization.

All inference algorithms should return an instance of one of these classes.
"""

import json
from collections import Counter
from collections.abc import Iterable
import itertools

import numpy as np
import networkx as nx

MODULE_PATH = __file__[: -len("source_results.py")]
MODULE_PATH = (
    MODULE_PATH
    if len(MODULE_PATH) > 0 and (MODULE_PATH[-1] == "/" or MODULE_PATH[-1] == "\\")
    else MODULE_PATH + "/"
)


def node_set_distance(s1, s2, G):
    """Implements a distance measure between vertex sets (of possibly different sizes).

    Parameters
    ----------
    s1 : array-like
        first vertex set
    s2 : array-like
        second vertex set
    G : NetworkX Graph
        graph to search on
    """
    perm_scores = {}

    if isinstance(s1, Iterable):
        s1 = list(s1)
    else:
        s1 = [s1]
    if isinstance(s2, Iterable):
        s2 = list(s2)
    else:
        s2 = [s2]

    for s2_perm in itertools.permutations(s2):
        perm_scores[s2_perm] = 0
        for i in range(min(len(s1), len(s2))):
            perm_scores[s2_perm] += nx.shortest_path_length(
                G, source=s1[i], target=s2_perm[i]
            )
        if len(s1) > len(s2):
            for j in range(i, len(s1)):
                min_add = np.inf
                for s in s2_perm:
                    d = nx.shortest_path_length(G, source=s1[j], target=s)
                    if d < min_add:
                        min_add = d
                perm_scores[s2_perm] += min_add
        if len(s2) > len(s1):
            for j in range(i, len(s2_perm)):
                min_add = np.inf
                for s in s1:
                    d = nx.shortest_path_length(G, source=s2_perm[j], target=s)
                    if d < min_add:
                        min_add = d
                perm_scores[s2_perm] += min_add
    return min(perm_scores.values())


class SourceResult:
    """Abstract class outlining response object for the result of a source inference algorithm.

    Parameters
    ----------
    source_type : str
        either "single-source" or "multi-source"
    inference_method : str
        name of the source localization algorithm used
    scores : dict
        per-item scores for ranking, retrieval, etc.
    G : NetworkX Graph
        The original graph the infection process was run on.
        I is a subgraph of G induced by infected vertices at observation time.
    algorithm_details : bool
        if True, includes relevant information about the source
        inference algorithm used
    reverse : bool (default True)'
        if True, ranks items from highest score to lowest
        if False, ranks items from lowest score to highest
    """

    def __init__(
        self,
        source_type,
        inference_method,
        scores,
        G,
        algorithm_details=True,
        reverse=True,
    ):
        """Abstract class outlining response object for the result of a source inference algorithm.

        Parameters
        ----------
        source_type : str
            either "single-source" or "multi-source"
        inference_method : str
            name of the source localization algorithm used
        scores : dict
            per-item scores for ranking, retrieval, etc.
        G : NetworkX Graph
            The original graph the infection process was run on.
            I is a subgraph of G induced by infected vertices at observation time.
        algorithm_details : bool
            if True, includes relevant information about the source
            inference algorithm used
        reverse : bool (default True)'
            if True, ranks items from highest score to lowest
            if False, ranks items from lowest score to highest
        """
        if not isinstance(G, nx.classes.graph.Graph):
            raise ValueError("G must be a NetworkX graph.")
        else:
            self.G = G

        source_type = source_type.lower()
        if source_type not in ["single-source", "multi-source"]:
            raise ValueError("Source type must be single- or multi-source.")
        self.data = {
            "scores": scores,
            "inference method": {"name": inference_method, "source_type": source_type},
            "G": G,
        }
        if algorithm_details:
            algorithms = json.load(open(MODULE_PATH + "algorithm_details.json"))
            for k in algorithms[source_type][inference_method]:
                self.data["inference method"][k] = algorithms[source_type][
                    inference_method
                ][k]
        self.reverse = reverse
        return None

    def rank(self):
        """Rank nodes by score.

        Returns
        -------
        list of item indices
        """
        scores = self.data["scores"]
        return sorted(scores, key=scores.get, reverse=self.reverse)

    def topn(self, n=1):
        """Returns the top n item indices by rank.

        Rank can be highest-first (reverse==True) or lowest-first (reverse==False)

        Parameters
        ----------
        n : int
            number of item indices to return

        Returns
        -------
        list of item indices
        """
        if not isinstance(n, int):
            raise ValueError("n must be an integer.")
        rank = self.rank()
        return rank[:n]

    def evaluate_solution_rank(self, true_source):
        """Finds the rank of the true source, by the algorithm's scoring protocol.

        Parameters
        ----------
        true_source : graph index - str, int, etc.
            the actual source node
        """
        single_source = len(self.topn(n=1)) == 1
        if isinstance(true_source, (list, tuple)) and len(true_source) == 1:
            true_source = true_source[0]

        return self.get_rank(true_source, soft_rank=True)

    def evaluate_distance(self, true_source):
        """Finds the shortest path length between each node in the solution set and the
        true souce.

        Parameters
        ----------
        true_source : tuple
            the actual source set
        """
        eval_scores = {h: np.inf for h in self.data["scores"]}
        for s in eval_scores.keys():
            eval_scores[s] = node_set_distance(G=self.G, s1=s, s2=true_source)
        return eval_scores

    def evaluate(self, true_source):
        """Runs evaluation algorithms and returns a dictionary of results.

        Parameters
        ----------
        true_source : graph index - str, int, etc.
            the actual source node
        """
        dist = self.evaluate_distance(true_source=true_source)
        top_sol = self.topn(n=1)[0]
        rank = self.evaluate_solution_rank(true_source=true_source)
        evaluation_results = {
            "true source": true_source,
            "distance": {
                "top score's distance": {top_sol: dist[top_sol]},
                "all distances": dist,
            },
            "rank": rank,
            "rank %": rank / len(self.data["scores"]),
        }
        return evaluation_results


class SingleSourceResult(SourceResult):
    """Response object for the result of single-source inference.

    Parameters
    ----------
    inference_method : str
        name of the source localization algorithm used
    scores : dict
        per-node scores for ranking, retrieval, etc.
    algorithm_details : bool
        if True, includes relevant information about the source
        inference algorithm used
    reverse : bool (default True)'
        if True, ranks items from highest score to lowest
        if False, ranks items from lowest score to highest
    """

    def __init__(self, *args, **kwargs):
        """Response object for the result of single-source inference.

        Parameters
        ----------
        inference_method : str
            name of the source localization algorithm used
        scores : dict
            per-node scores for ranking, retrieval, etc.
        algorithm_details : bool
            if True, includes relevant information about the source
            inference algorithm used
        reverse : bool (default True)'
            if True, ranks items from highest score to lowest
            if False, ranks items from lowest score to highest
        """
        super().__init__(*args, **kwargs)
        return None

    def get_rank(self, v, soft_rank=False):
        """Returns the rank of vertex (1 = "best")

        Parameters
        ----------
        v : graph index - str, int, etc.
            vertex of interest
        soft_rank : bool
            if True and v is not in the list of hypotheses, returns 1 more
            than the number of hypotheses

        Returns
        -------
        int
        """
        rank = self.rank()

        if soft_rank and v not in rank:
            return len(rank) + 1
        return rank.index(v) + 1


class MultiSourceResult(SourceResult):
    """Response object for the result of mutli-source inference.

    Parameters
    ----------
    inference_method : str
        name of the source localization algorithm used
    scores : dict
        per-item scores for ranking, retrieval, etc.
    algorithm_details : bool
        if True, includes relevant information about the source
        inference algorithm used
    reverse : bool (default True)'
        if True, ranks items from highest score to lowest
        if False, ranks items from lowest score to highest
    """

    def __init__(self, *args, **kwargs):
        """Response object for the result of mutli-source inference.

        Parameters
        ----------
        inference_method : str
            name of the source localization algorithm used
        scores : dict
            per-item scores for ranking, retrieval, etc.
        algorithm_details : bool
            if True, includes relevant information about the source
            inference algorithm used
        reverse : bool (default True)'
            if True, ranks items from highest score to lowest
            if False, ranks items from lowest score to highest
        """
        super().__init__(*args, **kwargs)
        return None

    def get_rank(self, s, soft_rank=False):
        """Returns the rank of the provided node set (1 = "best")

        Parameters
        ----------
        s : list
            node set of graph indices
        soft_rank : bool
            if True and v is not in the list of hypotheses, returns 1 more
            than the number of hypotheses

        Returns
        -------
        int
        """
        rank = self.rank()
        r = 1
        for q in rank:
            if Counter(s) == Counter(q):
                break
            r += 1
        if not soft_rank and r > len(self.data["scores"]):
            raise ValueError("Proposed source set not found among top hypotheses.")
        return r



import numpy as np
import torch_geometric.datasets as ds
import random
import ndlib.models.epidemics as ep
import ndlib.models.ModelConfig as mc

def random(seed):
    random.seed(seed)
    np.random.seed(seed)
    return seed

# todo
def degree():
    return 

# todo
def eigen():
    return 


import itertools

import networkx as nx
import numpy as np

from ..source_results import MultiSourceResult
from ...utils import estimators
from .. import single_source


def fast_multisource_lisn(I, G, t, number_sources=None):
    """Greedily runs single-source LISN algorithm on each estimated infection
    subgraph attributable to each of the hypothesized number of sources.

    Parameters
    ----------
    I : NetworkX Graph
        The infection subgraph observed at a particular time step
    G : NetworkX Graph
        The original graph the infection process was run on.
        I is a subgraph of G induced by infected vertices at observation time.
    t : int
        the observation timestep corresponding to I
    number_sources : int or None (optional)
        if int, this is the hypothesized number of infection sources
        if None, estimates the number of sources

    Notes
    -----
    The Jordan infection center is the vertex with minimum infection eccentricity.
    This is described in [1]_ and [2]_.

    Examples
    --------
    >>> result = cosasi.multiple_source.fast_multisource_jordan_centrality(I, G)

    References
    ----------
    .. [1] L. Ying and K. Zhu,
        "On the Universality of Jordan Centers for Estimating Infection Sources in Tree Networks"
        IEEE Transactions of Information Theory, 2017
    .. [2] L. Ying and K. Zhu,
        "Diffusion Source Localization in Large Networks"
        Synthesis Lectures on Communication Networks, 2018
    """
    if not number_sources:
        number_sources, subgraphs = estimators.number_sources(
            I, return_source_subgraphs=True
        )
    else:
        number_sources, subgraphs = estimators.number_sources(
            I, number_sources=number_sources, return_source_subgraphs=True
        )

    sources_scores = [
        {
            k: v
            for k, v in single_source.lisn(I=subgraphs[i], G=G, t=t)
            .data["scores"]
            .items()
            if v != -np.inf
        }
        for i in range(number_sources)
    ]
    data = [list(d.keys()) for d in sources_scores]
    product_scores = {}

    for item in itertools.product(*data):
        idx = tuple(item)
        product_scores[idx] = 0
        for i in range(len(idx)):
            product_scores[idx] += sources_scores[i][idx[i]]
    result = MultiSourceResult(
        source_type="multi-source",
        inference_method="fast multi-source lisn",
        scores=product_scores,
        G=G,
    )
    return result



import xflow_loader

from xflow.dataset.nx import BA, connSW
from xflow.dataset.pyg import Cora
from xflow.diffusion import SI, IC, LT
from xflow.seed import random as seed_random, degree as seed_degree, eigen as seed_eigen
from xflow.util import run

# graphs to test
fn = lambda: connSW(n=1000, beta=0.1)
fn.__name__ = 'connSW'
gs = [fn, BA]

# diffusion models to test
# TODO actually, no need to import in this main.py, because the diffusion models are embeded in the methods
df = [SI, IC, LT]

# configurations of IM experiments
from xflow.method.im import pi as im_pi, degree as im_degree, sigma as im_sigma, eigen as im_eigen, celf as im_celf,celfpp as im_celfpp, greedy as im_greedy
me = [im_pi, im_eigen]
rt = run (
    graph = gs, diffusion = df, 
    method = me, eval = 'im', epoch = 10, 
    budget = 10, 
    output = [ 'animation', 'csv', 'fig'],
    seeds = seed_random)

# configurations of IBM experiments
from xflow.method.ibm import pi as ibm_pi, degree as ibm_degree, sigma as ibm_sigma, eigen as im_eigen, greedy as ibm_greedy
me = [ibm_pi, ibm_sigma, ibm_degree]
rt = run (
    graph = gs, diffusion = df, 
    method = me, eval = 'ibm', epoch = 10,
    budget = 10,
    output = [ 'animation', 'csv', 'fig'],
    seeds = seed_random)
 
# configurations of SL experiments
from xflow.method.cosasi.source_inference.multiple_source.netsleuth import netsleuth, fast_multisource_netsleuth
from xflow.method.cosasi.source_inference.multiple_source.lisn import fast_multisource_lisn
from xflow.method.cosasi.source_inference.multiple_source.jordan import fast_multisource_jordan_centrality

me = [netsleuth]
rt = run (
    graph = gs, diffusion = df, 
    method = me, eval = 'sl', epoch = 10,
    budget = 10,
    output = [ 'animation', 'csv', 'fig'],
    seeds = seed_random)



import random
import os, sys
import json

sys.path.insert(0, os.getcwd())

import numpy as np
import networkx as nx

import cosasi

MODULE_PATH = __file__[: -len("benchmark.py")]
MODULE_PATH = (
    MODULE_PATH
    if len(MODULE_PATH) > 0 and (MODULE_PATH[-1] == "/" or MODULE_PATH[-1] == "\\")
    else MODULE_PATH + "/"
)
ALGORITHMS_PATH = MODULE_PATH[: -len("benchmark/")] + "source_inference/"


def _get_relevant_namespaces(
    source_type=None, information_type="single snapshot", epidemic_model=None
):
    """Retrieves the functional names of all applicable source inference algorithms.

    Parameters
    ----------
    source_type : str or None (optional)
        one of None, "single-source", or "multi-source"
        If None, we consider any source type
    information_type : str
        describes the information the source inference algorithm receives
        e.g. "single snapshot"
    epidemic_model : str or None (optional)
        specifies the epidemic model, e.g. SI, SIS, SIR
        if None, ignores this constraint
    """
    valid_namespaces = []
    algorithms_dict = json.load(open(ALGORITHMS_PATH + "algorithm_details.json"))

    if isinstance(source_type, type(None)):
        source_type_iter = list(algorithms_dict.keys())
    else:
        source_type_iter = [source_type]

    for source_type in source_type_iter:
        for alg_name in algorithms_dict[source_type]:
            if not isinstance(epidemic_model, type(None)):
                if (
                    epidemic_model.lower()
                    not in algorithms_dict[source_type][alg_name]["epidemic model"]
                ):
                    continue

            if (
                algorithms_dict[source_type][alg_name]["information type"]
                == information_type
                and algorithms_dict[source_type][alg_name]["status"] == "complete"
            ):
                valid_namespaces.append(
                    eval(algorithms_dict[source_type][alg_name]["namespace"])
                )
    return valid_namespaces


def _get_namespace_params(name, return_defaults=True):
    """Retrieves the names of the parameters and their default values.

    Parameters
    ----------
    name : function
        function namespace
    return_defaults : bool
        if True, also includes
    """
    arg_num = name.__code__.co_argcount
    param_names = name.__code__.co_varnames[:arg_num]

    if not return_defaults:
        return param_names

    params = {}
    if isinstance(name.__defaults__, type(None)):
        defaults = []
    else:
        defaults = list(name.__defaults__)[::-1]
    param_names = param_names[::-1]
    for i in range(len(param_names)):
        if i < len(defaults):
            arg = defaults[i]
        else:
            arg = ""
        params[param_names[i]] = arg
    return params


def _execute_algorithm_from_namespace(name, what_we_know):
    """Runs a source inference algorithm, passing what we know as arguments.

    Parameters
    ----------
    name : function
        function namespace
    what_we_know : dict
        dictionary of arguments we want to pass to the algorithm
    """
    function_args = _get_namespace_params(name=name, return_defaults=True)

    for param in what_we_know:
        if param in function_args:
            function_args[param] = what_we_know[param]

    if "" in function_args.values():
        raise ValueError(
            "Insufficient arguments provided.",
            function_args,
            what_we_know,
            name,
            name.__defaults__,
        )

    return name(**function_args)


class BenchmarkFromDetails:
    """Benchmarking tool using provided class args to pass to algorithms when available.

    Parameters
    ----------
    true_source : node or tuple of nodes
        the true source of the diffusion process
    G : NetworkX Graph
        The original graph the infection process was run on.
        I is a subgraph of G induced by infected vertices at observation time.
    information_type : str
        describes the information the source inference algorithm receives
        e.g. "single snapshot"
    I : NetworkX Graph
        The infection subgraph observed at a particular time step
    t : int
        the timestep corresponding to I
    observer_dict : dict or None (optional)
        takes a dict of observers and the timestamps at which they become infected.
    epidemic_model : str or None (optional)
        specifies the epidemic model, e.g. SI, SIS, SIR
        if None, ignores this constraint
    number_sources : int or None (optional)
        if int, this is the hypothesized number of infection sources
        if None, estimates the number of sources
    infection_rate : float or None (optional)
        Inter-node infection efficiency. If a float, must be in [0,1]
        if None, ignores this parameter
    """

    def __init__(
        self,
        true_source,
        G,
        information_type,
        I=None,
        t=None,
        observer_dict=None,
        epidemic_model=None,
        number_sources=None,
        infection_rate=None,
    ):
        """Benchmarking tool using provided class args to pass to algorithms when available.

        Parameters
        ----------
        true_source : node or tuple of nodes
            the true source of the diffusion process
        G : NetworkX Graph
            The original graph the infection process was run on.
            I is a subgraph of G induced by infected vertices at observation time.
        information_type : str
            describes the information the source inference algorithm receives
            e.g. "single snapshot"
        I : NetworkX Graph
            The infection subgraph observed at a particular time step
        t : int
            the timestep corresponding to I
        observer_dict : dict or None (optional)
            takes a dict of observers and the timestamps at which they become infected.
        epidemic_model : str or None (optional)
            specifies the epidemic model, e.g. SI, SIS, SIR
            if None, ignores this constraint
        number_sources : int or None (optional)
            if int, this is the hypothesized number of infection sources
            if None, estimates the number of sources
        infection_rate : float or None (optional)
            Inter-node infection efficiency. If a float, must be in [0,1]
            if None, ignores this parameter
        """
        self.epidemic_model = epidemic_model
        self.number_sources = number_sources
        self.information_type = information_type
        if isinstance(t, (int, float, type(None))):
            self.t = t
        else:
            raise ValueError("Time parameter must be an integer or float or None")
        self.observer_dict = observer_dict

        if information_type == "single snapshot" and (
            isinstance(I, type(None)) or isinstance(t, type(None))
        ):
            raise ValueError(
                "If information type is single snapshot, we need the infection subgraph and its corresponding timestep"
            )
        if information_type == "observers" and (isinstance(observer_dict, type(None))):
            raise ValueError(
                "If the information type is observers, we need the observer_dict"
            )

        if isinstance(G, nx.classes.graph.Graph):
            self.G = G
        else:
            raise ValueError("G must be a NetworkX graph.")

        if all(v in G for v in true_source):
            self.true_source = true_source
        elif true_source in G:
            self.true_source = true_source
        else:
            raise ValueError("All members of true_source must be in G.")

        if isinstance(I, (nx.classes.graph.Graph, type(None))):
            self.I = I
        else:
            raise ValueError("I must be a NetworkX graph.")

        if (
            isinstance(infection_rate, float) and 0.0 <= infection_rate <= 1.0
        ) or isinstance(infection_rate, type(None)):
            self.infection_rate = infection_rate
        else:
            raise ValueError("Infection rate must be a float between 0 and 1.")

        self.namespaces = self.get_namespaces()
        return None

    def get_namespaces(self):
        """Finds all source localization algorithms applicable to the contagion task
        specified in the class constructor.
        """
        if isinstance(self.number_sources, type(None)):
            source_type = None
        elif self.number_sources > 1:
            source_type = "multi-source"
        elif self.number_sources == 1:
            source_type = "single-source"
        else:
            raise NotImplementedError

        namespaces = _get_relevant_namespaces(
            source_type=source_type,
            information_type=self.information_type,
            epidemic_model=self.epidemic_model,
        )

        return namespaces

    def go(self):
        """Runs all available algorithms with the information we have on hand."""
        result_dict = {}
        what_we_know = {
            "G": self.G,
            "I": self.I,
            "observer_dict": self.observer_dict,
            "t": self.t,
            "number_sources": self.number_sources,
        }
        for alg in self.namespaces:
            result = _execute_algorithm_from_namespace(
                name=alg, what_we_know=what_we_know
            )
            inference_method = result.data["inference method"]["name"]
            source_type = result.data["inference method"]["source_type"]
            result_dict[source_type + " " + inference_method] = {
                "source result": result,
                "evaluation": result.evaluate(true_source=self.true_source),
            }

        return result_dict


class BenchmarkFromSimulation:
    """Benchmarking tool using provided simulation object to pass to algorithms when available.

    Parameters
    ----------
    contagion : cosasi.contagion.static_network_contagion.StaticNetworkContagion
        an already-run contagion object
    t : int
        the timestep corresponding to I
    information_type : str or None (optional)
        describes the information the source inference algorithm receives
        e.g. "single snapshot"
    observers : int or list
        If int, observers specifies the number of observation nodes
        If list, observers specifies the observation nodes directly
    """

    def __init__(self, contagion, t=None, information_type=None, observers=None):
        """Benchmarking tool using provided simulation object to pass to algorithms when available.

        Parameters
        ----------
        contagion : cosasi.contagion.static_network_contagion.StaticNetworkContagion
            an already-run contagion object
        t : int
            the timestep corresponding to I
        information_type : str or None (optional)
            describes the information the source inference algorithm receives
            e.g. "single snapshot"
        observers : int or list
            If int, observers specifies the number of observation nodes
            If list, observers specifies the observation nodes directly
        """
        true_source = contagion.get_source()
        if information_type == "single snapshot":
            if isinstance(t, type(None)):
                raise ValueError("If information type is snapshot, t is required")
            if not isinstance(t, int):
                raise ValueError("t must be an int")
            self.benchmarker = BenchmarkFromDetails(
                true_source=true_source,
                G=contagion.G,
                I=contagion.get_infected_subgraph(step=t),
                t=t,
                epidemic_model=contagion.model,
                number_sources=len(true_source),
                information_type=information_type,
                infection_rate=contagion.infection_rate,
            )
        elif information_type == "observers":
            if isinstance(observers, type(None)):
                raise ValueError(
                    "If information type is observers, the number of observers is required"
                )
            if not isinstance(observers, (int, list)):
                raise ValueError("observers must be an int or a list")
            self.benchmarker = BenchmarkFromDetails(
                true_source=true_source,
                G=contagion.G,
                observer_dict=contagion.get_observers(observers=observers),
                epidemic_model=contagion.model,
                number_sources=len(true_source),
                information_type=information_type,
                infection_rate=contagion.infection_rate,
            )
        else:
            raise NotImplementedError
        return None

    def go(self):
        """Runs all available algorithms with the information we have on hand."""
        return self.benchmarker.go()



# pip dependencies
#!pip install torch
#!pip install torch_geometric==2.2.0
#!pip install xflow-net==0.0.21
#!pip install networkx
#!pip install ndlib
#!pip install pyg_lib torch_scatter torch_sparse torch_cluster torch_spline_conv -f https://data.pyg.org/whl/torch-2.2.0+cpu.html
#!pip install torch_geometric_temporal

# group 1
import numpy as np
import random
import networkx as nx
from networkx import Graph
from torch_geometric.data.data import Data, torch
import xflow
from xflow.dataset.nx import connSW
import ndlib.models.epidemics as ep
import ndlib.models.ModelConfig as mc
import warnings
from torch_geometric.utils.convert import from_networkx
from sklearn.metrics import classification_report, ConfusionMatrixDisplay
import matplotlib.pyplot as plt

def connSW(n, beta=None):
    g = nx.connected_watts_strogatz_graph(n, 10, 0.1) #Generate connSW graph

    config = mc.Configuration()

    #add edge weights in ndlib config
    for a, b in g.edges():
        weight = random.randrange(40,80)
        weight = round(weight / 100, 2)
        if beta:
            weight = beta
        g[a][b]['weight'] = weight
        config.add_edge_configuration("threshold", (a, b), weight)

    return g, config

def BA(n=1000, beta=None):
    g = nx.barabasi_albert_graph(n, 5)

    config = mc.Configuration()

    for a, b in g.edges():
        weight = random.randrange(40,80)
        weight = round(weight / 100, 2)
        if beta:
            weight = beta
        g[a][b]['weight'] = weight
        config.add_edge_configuration("threshold", (a, b), weight)

    return g, config

def ER(n=5000, beta=None):
    g = nx.erdos_renyi_graph(n, 0.1)

    while nx.is_connected(g) == False:
        g = nx.erdos_renyi_graph(n, 0.1)

    config = mc.Configuration()

    for a, b in g.edges():
        weight = random.randrange(40,80)
        weight = round(weight / 100, 2)
        if beta:
            weight = beta
        config.add_edge_configuration("threshold", (a, b), weight)
        g[a][b]['weight'] = weight

    return g, config

graph_gen_dict = {'connSW':connSW, 'BA':BA, 'ER':ER} #dictionary of string to graph gen mappings, shpuld be a generator that
                                                     #returns g as networkx grapg and config as ndlib config
                                                     #and accepts values of n (nodes) and beta (edge weight override)

def setup(graph_kind, graph_size, graph_beta, inf_beta, inf_gamma, inf_initial_frac):

    #generate a graph
    for name, gen_function in graph_gen_dict.items():
        if name == graph_kind:
            (g, config) = gen_function(n=graph_size, beta=graph_beta)
            break
    #make sure graph gen was sucessful
    if g is None:
        raise Exception('Graph generation function not known.')

    #sir model setup
    if inf_beta is None:
        inf_beta = random.uniform(0.01,0.06)
    if inf_gamma is None:
        inf_gamma = random.uniform(0.005,0.03)
    if inf_initial_frac is None:
        inf_initial_frac = random.uniform(0.02,0.05)

    config.add_model_parameter('beta', inf_beta)
    config.add_model_parameter('gamma', inf_gamma)
    config.add_model_parameter("fraction_infected", inf_initial_frac)

    model = ep.SIRModel(g)
    model.set_initial_status(config)

    return g,model,config

def run_sim(distance, interval_lower, g, model):
    #choose timesteps, dont run longer than needed
    diameter = nx.diameter(g)
    dist_max = max(distance)
    r = interval_lower
    if r < 0:
        if (diameter-dist_max <0):
            warnings.warn(f'Distance of {dist_max} exceeds graph diameter of {diameter}.')
            diameter = dist_max
        if (diameter-dist_max ==0):
            r=0
        else:
            r = random.randrange(diameter-dist_max) #choose random timesteps


    #run the simulation
    iterations = model.iteration_bunch(r+dist_max+1, node_status=True)

    node_states_iterations = []
    node_states = {}
    for iteration in iterations:
        if iteration['iteration'] == 0:
            node_states = {node: status for node, status in iteration['status'].items()}
        else:
            for node, status in iteration['status'].items():
                node_states[node] = status
        node_states_iterations.append(node_states.copy())
    #print(node_states_iterations)
    return node_states_iterations


def format_sim_result(intervals, iterations, obs_type, g):
    simulation_result = []
    if obs_type == 'numpy':
        #get result at each timestep
        for interval in intervals:
            node_states = iterations[interval]

            obs = np.array(list(node_states.values()))
            #build the snapshot using the observation at this iteration
            snapshot = {
                'time' : interval,
                'observation' : obs
            }
            #append the snapshot to sim result
            simulation_result.append(snapshot)

        return simulation_result


    if obs_type == 'torch' or obs_type == 'networkx':
        start = intervals[0]
        node_states = iterations[start]
        #node_states = {node: status for node, status in iteration['status'].items()}
        nx.set_node_attributes(g, node_states, name = 'state_x')
        nx.set_node_attributes(g, start, name = 'time_x')

        #get result at each timestep
        simulation_result = []
        for interval in intervals[1:]:
            node_states = iterations[interval]
            #node_states = {node: status for node, status in iteration['status'].items()}

            if obs_type == 'networkx':
                nx.set_node_attributes(g, node_states, name = 'state_y')
                nx.set_node_attributes(g, interval, name = 'time_y')
                obs = g.copy()
            elif obs_type == 'torch':
                nx.set_node_attributes(g, interval, name = 'time_y')
                obs = from_networkx(g.copy(), group_node_attrs=['state_x','time_x','time_y'], group_edge_attrs=['weight'])
                state = np.array(list(node_states.values()))
                obs.y = torch.tensor(state)


            #append the snapshot to sim result
            simulation_result.append(obs)
        return simulation_result

    raise Exception("Acceptable values for obs_storage are: 'numpy', 'networkx', 'torch' ")

def forward(distance,
            num_results=10,
            obs_type = 'numpy',
            graph_kind = 'connSW',
            graph_size=1000,
            graph_beta = None,
            inf_beta = None,
            inf_gamma = None,
            inf_initial_frac = None,
            interval_lower = -1,
            ):

    if isinstance(distance,int):
        distance = [distance]

    #generate results
    results = []
    for n in range(num_results):
        (g,inf_model, sir_config) = setup(graph_kind, graph_size, graph_beta, inf_beta, inf_gamma, inf_initial_frac)
        iterations = run_sim(distance, interval_lower, g, inf_model)

        intervals =[len(iterations) -max(distance)-1]
        for d in distance:
            intervals.append(intervals[0] + d)

        simulation_result = format_sim_result(intervals, iterations, obs_type, g.copy())

        #append two results that are distance apart
        results.append({'observations': simulation_result,
                        'base_graph': g,
                        'SIR_config': sir_config})


    return results

def backward(distance,
            num_results=10,
            obs_type = 'numpy',
            graph_kind = 'connSW',
            graph_size=1000,
            graph_beta = None,
            inf_beta = None,
            inf_gamma = None,
            inf_initial_frac = None,
            interval_lower = -1,
            ):

    if isinstance(distance,int):
        distance = [distance]

    #generate results
    results = []
    for n in range(num_results):
        (g,inf_model, sir_config) = setup(graph_kind, graph_size, graph_beta, inf_beta, inf_gamma, inf_initial_frac)
        iterations = run_sim(distance, interval_lower, g, inf_model)

        intervals =[len(iterations)-1]
        for d in distance:
            intervals.append(intervals[0] - d)

        simulation_result = format_sim_result(intervals, iterations, obs_type, g.copy())

        #append observations, the starting graph, the starting SIR model
        results.append({'observations': simulation_result,
                        'base_graph': g,
                        'SIR_config': sir_config})

    return results


def graph_eval(obs_true, obs_pred, display=False):
    from sklearn.metrics import classification_report, ConfusionMatrixDisplay
    import matplotlib.pyplot as plt

    if not isinstance(obs_true, np.ndarray) or not isinstance(obs_pred, np.ndarray):
        raise Exception('Observation inputs must be numpy arrays')


    cr = classification_report(obs_true, obs_pred, zero_division=0)

    if display:
        ConfusionMatrixDisplay.from_predictions(obs_true, obs_pred)
        plt.show()
        print(cr)

    return cr




import math
import random

import networkx as nx

from ...utils import list_product
from ..source_results import SingleSourceResult


def rumor_centrality_root(I, v, return_all_values=True):
    """Computes rumor centrality for all nodes, assuming a spanning tree rooted at v.

    Parameters
    ----------
    I : NetworkX Graph
        The infection subgraph observed at a particular time step
    v : graph index - str, int, etc.
        The vertex rooting
    return_all_values : bool
        Specifies whether you want the full rumor centrality dict.
        If False, returns only the value for node v

    Notes
    -----
    Rumor centrality was introduced in the seminal work [1]_. This is a more "literal"
    interpretation of their algorithm. `rumor_centrality` averages these results over all
    possible BFS rooting schemes.

    References
    ----------
    .. [1] S. Devavrat and T. Zaman,
        "Rumors in a network: Who's the culprit?."
        IEEE Transactions on Informatidon Theory, 2011
        https://devavrat.mit.edu/wp-content/uploads/2017/10/Rumors-in-a-network-whos-the-culprit.pdf
    """
    N = len(I)
    G = nx.bfs_tree(I, v)

    # sort nodes by depth from leaves to root
    depths = nx.shortest_path_length(G, v)
    nodes_by_depth = sorted(depths, key=depths.get, reverse=True)

    # message-passing data objects; indexing is dict[destination][source]
    t = {
        i: {j: 0 for j in nodes_by_depth} for i in nodes_by_depth
    }  # subtree size messages
    p = {
        i: {j: 0 for j in nodes_by_depth} for i in nodes_by_depth
    }  # subtree product messages
    r = {i: 0 for i in nodes_by_depth}  # rumor centrality values

    for u in nodes_by_depth:
        children_u = [e[1] for e in G.out_edges(u)]
        if u != v:
            parent_u = list(G.in_edges(u))[0][0]
        if G.out_degree(u) == 0:
            # u is a leaf
            if "parent_u" in locals():
                t[parent_u][u] = 1
                p[parent_u][u] = 1
        else:
            if u != v:
                # u is not root
                if "parent_u" in locals():
                    t[parent_u][u] = 1 + sum([t[u][j] for j in children_u])
                    p[parent_u][u] = t[parent_u][u] * list_product(
                        [p[u][j] for j in children_u]
                    )
    for u in nodes_by_depth[::-1]:
        children_u = [e[1] for e in G.out_edges(u)]
        if u == v:
            # u is root
            r[u] = math.factorial(N) / list_product([p[u][j] for j in children_u])
        else:
            parent_u = list(G.in_edges(u))[0][0]
            r[u] = r[parent_u] * t[parent_u][u] / (N - t[parent_u][u])
    for u in nodes_by_depth:
        r[u] /= len(G)
    if not return_all_values:
        return r[v]
    return r


def rumor_centrality(I, G=None, v=None, normalize=True, only_roots=False):
    """Computes rumor centrality for all nodes in G.

    Parameters
    ----------
    I : NetworkX Graph
        The infection subgraph observed at a particular time step
    G : NetworkX Graph (optional)
        The original graph the infection process was run on.
        I is a subgraph of G induced by infected vertices at observation time.
        This is not actually used, but exists to match the format of other algorithms
    v : graph index - str, int, etc. (optional)
        if provided, returns the rumor centrality of v only.
    normalize : bool
        If True, scales all rumor centrality values to between 0 and 1
    only_roots : bool
        Aggregation strategy, as we compute rumor_centrality_root over all possible
        root nodes.
        If True, we only keep the rumor_centrality_root value for the root node
        If False, we keep the rumor_centrality_root values for all nodes

    Notes
    -----
    Rumor centrality was introduced in the seminal work [1]_. `rumor_centrality_root` is a
    more "literal" interpretation of their algorithm. `rumor_centrality` (this function)
    averages these results over all possible BFS rooting schemes.


    Examples
    --------
    >>> result = cosasi.single_source.rumor_centrality(I, G)

    References
    ----------
    .. [1] S., Devavrat and T. Zaman,
        "Rumors in a network: Who's the culprit?."
        IEEE Transactions on Information Theory, 2011
        https://devavrat.mit.edu/wp-content/uploads/2017/10/Rumors-in-a-network-whos-the-culprit.pdf
    """
    if v and v not in I:
        raise ValueError("Provided node is not in I.")

    # iterate over possible roots, and average over spanning trees
    rumor_centrality_dict = {i: 0 for i in I.nodes}
    for root in rumor_centrality_dict:
        if only_roots:
            rumor_centrality_dict[root] = rumor_centrality_root(
                I, root, return_all_values=False
            )
        else:
            r = rumor_centrality_root(I, root, return_all_values=True)
            for node in I.nodes:
                if node in r:
                    rumor_centrality_dict[node] += r[node]
    for node in rumor_centrality_dict:
        rumor_centrality_dict[node] /= len(I)
    if normalize:
        max_val = max(rumor_centrality_dict.values())
        for node in rumor_centrality_dict:
            rumor_centrality_dict[node] /= max_val
    if v:
        return rumor_centrality_dict[v]
    result = SingleSourceResult(
        source_type="single-source",
        inference_method="rumor centrality",
        scores=rumor_centrality_dict,
        G=G,
    )
    return result



import itertools

import networkx as nx
import numpy as np

from ..source_results import MultiSourceResult
from ...utils import estimators
from .. import single_source


def netsleuth(I, G, hypotheses_per_step=1):
    """Implements the multi-source NETSLEUTH algorithm to score combinations
    of nodes in G.

    Parameters
    ----------
    I : NetworkX Graph
        The infection subgraph observed at a particular time step
    G : NetworkX Graph
        The original graph the infection process was run on.
        I is a subgraph of G induced by infected vertices at observation time.
    hypotheses_per_step : int (default 1)
        number of candidate sources to be kept per iteration of NETSLEUTH.
        Particular usage is described in greater detail in `Notes` section.

    Notes
    -----
    The number of source hypotheses returned will be hypotheses_per_step*[number of seed nodes],
    the latter of which is automatically determined via minimum description length
    calculations.

    NETSLEUTH is described in [1]_ and [2]_.

    NETSLEUTH has linear complexity with the number of edges of the infected subgraph,
    edges of the frontier set, and vertices of the infected subgraph.

    The standard n-source version of NETSLEUTH operates as follows:

    1. Obtain Source 1 via single-source method

    2. Delete Source 1 from infection subgraph; obtain Source 2 via single-source method

        ...

    n. Delete Source n-1 from infection subgraph; obtain Source n via single-source method.

    This does not lend itself to ranking alternative hypotheses, so we implement a
    more general variant:

    1. Obtain top ``hypotheses_per_step``-many candidates for Source 1 via single-source
    method; each corresponds to one hypothesis source set, each of size 1

    2. For each hypothesis source set, delete these nodes from a copy of the infection subgraph,
    then obtain top ``hypotheses_per_step``-many candidates for Source 2 via single-source
    method; construct ``|source sets| * hypotheses_per_step`` new source sets to replace the old
    source sets, each of size 2

        ...

    n. For each hypothesis source set, delete these nodes from a copy of the infection subgraph,
    then obtain top ``hypotheses_per_step``-many candidates for Source n via single-source
    method; construct |source sets|*``hypotheses_per_step`` new source sets to replace the old
    source sets, each of size n


    Examples
    --------
    >>> result = cosasi.multiple_source.netsleuth(I, G, number_sources=3, hypotheses_per_step=3)

    References
    ----------
    .. [1] B. Prakash, J. Vreeken, C. Faloutsos,
        "Spotting Culprits in Epidemics: How Many and Which Ones?"
        IEEE 12th International Conference on Data Mining, 2012
        https://ieeexplore.ieee.org/document/6413787
    .. [2] L. Ying and K. Zhu,
        "Diffusion Source Localization in Large Networks"
        Synthesis Lectures on Communication Networks, 2018
    """
    multisource_scores = {}
    mdl_decreasing = True
    this_mdl = np.inf
    last_mdl = np.inf
    i = 1

    while mdl_decreasing:
        if i == 1:
            step_result = single_source.netsleuth(I, G)
            for s in step_result.topn(hypotheses_per_step):
                multisource_scores[(s)] = estimators.description_length([s], G)
        else:
            new_multisource_scores = {}
            for j in multisource_scores.keys():
                H = I.copy()
                if i == 2:
                    H.remove_nodes_from([j])
                else:
                    H.remove_nodes_from(j)
                step_result = single_source.netsleuth(H, G)
                for s in step_result.topn(hypotheses_per_step):
                    if i == 2:
                        new_s = tuple([j] + [s])
                    else:
                        new_s = tuple(list(j) + [s])
                    new_multisource_scores[new_s] = estimators.description_length(
                        list(new_s), G
                    )
            multisource_scores = new_multisource_scores

        # update mdl tracker
        last_mdl = this_mdl
        this_mdl = min(multisource_scores.values())
        mdl_decreasing = this_mdl < last_mdl
        i += 1

    result = MultiSourceResult(
        source_type="multi-source",
        inference_method="netsleuth",
        scores=multisource_scores,
        G=G,
        reverse=False,
    )
    return result


def fast_multisource_netsleuth(I, G, number_sources=None):
    """Greedily runs single-source NETSLEUTH on each estimated infection subgraph attributable
    to each of the hypothesized number of sources.

    Parameters
    ----------
    I : NetworkX Graph
        The infection subgraph observed at a particular time step
    G : NetworkX Graph
        The original graph the infection process was run on.
        I is a subgraph of G induced by infected vertices at observation time.
    number_sources : int or None (optional)
        if int, this is the hypothesized number of infection sources
        if None, estimates the number of sources

    Examples
    --------
    >>> result = cosasi.multiple_source.fast_multisource_netsleuth(I, G)

    Notes
    -----
    Unofficial variant of multisource NETSLEUTH intended for fast computation and ranking,
    because the typical multisource version does not lend itself to scoring many possible
    source sets.

    NETSLEUTH is described in [1]_ and [2]_. More authoritative implementation is found in
    `multisource.netsleuth`.

    References
    ----------
    .. [1] B. Prakash, J. Vreeken, C. Faloutsos,
        "Spotting Culprits in Epidemics: How Many and Which Ones?"
        IEEE 12th International Conference on Data Mining, 2012
        https://ieeexplore.ieee.org/document/6413787
    .. [2] L. Ying and K. Zhu,
        "Diffusion Source Localization in Large Networks"
        Synthesis Lectures on Communication Networks, 2018
    """
    if not number_sources:
        number_sources, subgraphs = estimators.number_sources(
            I, return_source_subgraphs=True
        )
    else:
        number_sources, subgraphs = estimators.number_sources(
            I, number_sources=number_sources, return_source_subgraphs=True
        )

    sources_scores = [
        {
            k: v
            for k, v in single_source.netsleuth(subgraphs[i], G).data["scores"].items()
            if v != -np.inf
        }
        for i in range(number_sources)
    ]
    data = [list(d.keys()) for d in sources_scores]
    product_scores = {}

    for item in itertools.product(*data):
        idx = tuple(item)
        product_scores[idx] = 0
        for i in range(len(idx)):
            product_scores[idx] += sources_scores[i][idx[i]]
    result = MultiSourceResult(
        source_type="multi-source",
        inference_method="fast multi-source netsleuth",
        scores=product_scores,
        G=G,
    )
    return result
